import requests
from loguru import logger
from sybil_engine.config.app_config import get_cex_conf, get_cex_data
from sybil_engine.domain.balance.balance import NativeBalance, Erc20Balance
from sybil_engine.domain.balance.balance_utils import get_native_balance
from sybil_engine.domain.balance.tokens import Erc20Token
from sybil_engine.domain.cex.okx import OKX

total_zro = 0

def claimer_zro(account, chain_instance, token, web3):
    allocation_data = get_allocation(account.address)
    if not allocation_data:
        raise Exception("Failed to fetch allocation data.")
    amount_raw = int(allocation_data.get('zroAllocation', {}).get('asBigInt'))
    logger.info(f"ZRO Allocation for {account.address} is {amount_raw}")
    if token == 'NATIVE':
        withdraw_token = chain_instance['gas_token']
        claim_price = NativeBalance(int((round(amount_raw / 10 ** 18) + 1) * 0.00005 * 10 ** 18),
                                    chain_instance['chain'], token)
        balance = get_native_balance(account, web3, chain_instance)
        logger.info(f"Native balance is {balance}")
    else:
        claim_price = Erc20Balance(
            int((round(amount_raw / 10 ** 18) + 1) * 0.00003 * 10 ** 18),
            chain_instance['chain'],
            token,
            decimal=18
        )
        withdraw_token = token
        balance = Erc20Token(chain_instance['chain'], token, web3).balance(account)
        logger.info(f"Native balance is {balance}")
    logger.info(f"Need {claim_price} for claim")
    if claim_price.wei > balance.wei:
        password, cex_data = get_cex_data()
        cex_obj = OKX(cex_data[get_cex_conf()], password)
        cex_obj.withdrawal(account.address, chain_instance['chain'], float(balance.readable()), withdraw_token)

    proof = get_proof(account.address)
    logger.info(proof)
    global total_zro
    total_zro = total_zro + int(proof['amount'])
    amount_to_claim = int(proof['amount'])
    merkle_proof = proof['proof'].split('|')
    #contract_address = get_contracts_for_chain(chain_instance['chain'])['ZRO_CLAIMER']
    #claimer = L0Claim(contract_address, web3)
    #claimer.donateAndClaim(account, claim_price.wei, amount_to_claim, merkle_proof)

def get_proof(wallet):
    url = f"https://www.layerzero.foundation/api/proof/{wallet}?address={wallet}"
    response = requests.get(url)
    if response.status_code == 200:
        return response.json()
    else:
        return None


def get_allocation(wallet):
    url = f"https://www.layerzero.foundation/api/allocation/{wallet}"
    response = requests.get(url)
    if response.status_code == 200:
        return response.json()
    else:
        return None


from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/zerolend.json")


class ZeroLendContract(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def deposit_eth(self, account, amount):
        sender = account.address
        txn_params = self.build_generic_data(sender, False)

        txn_params['value'] = amount

        return self.contract.functions.depositETH(
            '0x2f9bB73a8e98793e26Cb2F6C4ad037BDf1C6B269',
            account.address,
            0
        ).build_transaction(txn_params)

    @evm_transaction
    def withdraw_eth(self, account, amount):
        sender = account.address
        txn_params = self.build_generic_data(sender, False)

        return self.contract.functions.withdrawETH(
            '0x2f9bB73a8e98793e26Cb2F6C4ad037BDf1C6B269',
            int(amount * 0.98),
            account.address
        ).build_transaction(txn_params)



from datetime import datetime

from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/lineaswap_router.json")


class LineaSwapRouter(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def swap_exact_eth_for_tokens(self, account, amount_to_swap, token_in_address, token_out_address, amount_out_min):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)
        txn_params['value'] = amount_to_swap.wei if amount_to_swap.token == 'ETH' else 10000000000000

        paths = [token_in_address, token_out_address]
        deadline = int(datetime.now().timestamp() + 60 * 60 * 3)

        return self.contract.functions.swapExactETHForTokens(
            amount_out_min,
            paths,
            sender,
            deadline
        ).build_transaction(txn_params)

    @evm_transaction
    def swap_exact_tokens_for_eth(self, account, amount_to_swap, token_in_address, token_out_address, amount_out_min):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)

        paths = [token_in_address, token_out_address]
        deadline = int(datetime.now().timestamp() + 60 * 60 * 3)

        return self.contract.functions.swapExactTokensForETH(
            amount_to_swap.wei,
            amount_out_min,
            paths,
            sender,
            deadline
        ).build_transaction(txn_params)

    def get_amount_out(self, amount_to_swap, token_in_address, token_out_address):
        return self.contract.functions.getAmountsOut(
            amount_to_swap.wei,
            [token_in_address, token_out_address]
        ).call()[1]



import random

from loguru import logger
from sybil_engine.domain.balance.balance_utils import find_chain_with_max_usdc, find_chain_with_max_native
from sybil_engine.module.module import Module
from sybil_engine.utils.utils import randomized_sleeping
from sybil_engine.utils.web3_utils import print_account_data, get_all_account_data

from web3_wizzard_lib.core.modules.stargate.bridge import StargateBridge


class StargateAutoBridge(Module):
    module_name = 'STARGATE_AUTO_BRIDGE'
    module_config = 'stargate_auto_bridge_config'

    def execute(self, bridge_amount_interval, token, original_chain_sequence, sleep_interval, retry_interval,
                end_network, only_end_network_bridge, account):
        chain_sequence = original_chain_sequence.copy()

        logger.info(f"=============== Bridging for {account.address} ===============")

        account_data = get_all_account_data(account, chain_sequence)
        print_account_data(account_data)

        if token == 'USDC':
            max_usdc_chain, usdc, native_balance, web3 = find_chain_with_max_usdc(account_data)
        else:
            max_usdc_chain, usdc, native_balance, web3 = find_chain_with_max_native(account_data)

        logger.info(f"Chain with biggest {token} balance is {max_usdc_chain}")

        if only_end_network_bridge:
            if end_network == '':
                raise Exception("Env network should be declared on ONLY_END_NETWORK_BRIDGE=True")
            shuffled_chain_sequence = [max_usdc_chain] + [end_network]
        else:
            shuffled_chain_sequence = self.shuffle_chain_sequence(chain_sequence, max_usdc_chain, end_network)

        chain_sequence_readable = 'Result chain sequence: ' + ' -> '.join(shuffled_chain_sequence)
        logger.info(chain_sequence_readable)
        self.process_bridge(bridge_amount_interval, token, account, retry_interval, shuffled_chain_sequence,
                            sleep_interval)

    def process_bridge(self, bridge_amount_interval, bridge_token, account, retry_interval, shuffled_chain_sequence,
                       sleep_interval):
        for chain_number, _ in enumerate(shuffled_chain_sequence[:-1]):
            from_chain = shuffled_chain_sequence[chain_number]
            to_chain = shuffled_chain_sequence[chain_number + 1]

            StargateBridge(retry_interval).bridge(
                self.min_native_balance,
                account,
                bridge_amount_interval,
                bridge_token,
                from_chain,
                to_chain
            )
            randomized_sleeping(sleep_interval)

    def shuffle_chain_sequence(self, chain_sequence, max_usdc_chain, end_network):
        chain_sequence.remove(max_usdc_chain)

        while True:
            random.shuffle(chain_sequence)

            # If reshuffle doesn't result in the prohibited sequence, break out of the loop
            if not any((chain_sequence[i] == 'FANTOM' and chain_sequence[i + 1] == 'BASE') or
                       (chain_sequence[i] == 'BASE' and chain_sequence[i + 1] == 'FANTOM') or
                       (max_usdc_chain == 'FANTOM' and chain_sequence[0] == 'BASE') or
                       (max_usdc_chain == 'BASE' and chain_sequence[0] == 'FANTOM')
                       for i in range(len(chain_sequence) - 1)):
                break

        if end_network and chain_sequence[-1] != end_network:
            chain_sequence.append(end_network)

        return [max_usdc_chain] + chain_sequence

    def log(self):
        return "STARGATE AUTO"

    def parse_params(self, module_params):
        return (
            module_params['bridge_amount_interval'],
            module_params['bridge_token'],
            module_params['original_chain_sequence'],
            module_params['sleep_interval'],
            module_params['retry_interval'],
            module_params['end_network'],
            module_params['only_end_network_bridge']
        )



from sybil_engine.domain.dex import Dex

from web3_wizzard_lib.core.contract.lineaswaprouter import LineaSwapRouter


class LineaSwap(Dex):
    dex_name = 'lineaswap'
    swap_contract = "LINEASWAP"
    supported_chains = ['BASE', 'LINEA']

    def __init__(self, chain_instance, web3):
        super().__init__(chain_instance, web3)
        self.linea_swap_router = LineaSwapRouter(self.chain_contracts[self.swap_contract], self.web3)
        self.weth_address = self.tokens['WETH']

    def swap_native_for_token(self, account, amount_to_swap, slippage, to_token_address):
        amount_out_min = int(self.get_amount_out_min(amount_to_swap, self.weth_address, to_token_address) * slippage)
        args = [account, amount_to_swap, self.weth_address, to_token_address, amount_out_min]
        func = self.linea_swap_router.swap_exact_eth_for_tokens

        return args, func

    def swap_token_for_native(self, account, amount_to_swap, from_token_address, slippage):
        amount_out_min = int(self.get_amount_out_min(amount_to_swap, from_token_address, self.weth_address) * slippage)
        args = [account, amount_to_swap, from_token_address, self.weth_address, amount_out_min]
        func = self.linea_swap_router.swap_exact_tokens_for_eth

        return args, func

    def swap_token_for_token(self, account, amount_to_swap, slippage, from_token_address, to_token_address):
        raise Exception("Not supported")

    def get_amount_out_min(self, amount_to_swap, from_token_address, to_token_address):
        return self.linea_swap_router.get_amount_out(amount_to_swap, from_token_address, to_token_address)



from sybil_engine.contract.contract import Contract
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/scroll_oracle.json")

class ScrollOracle(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    def estimateCrossDomainMessageFee(self):
        return self.contract.functions.estimateCrossDomainMessageFee(168000).call()



import logging

from sybil_engine.module.module import Module
from sybil_engine.utils.utils import randomized_sleeping
from sybil_engine.utils.validation_utils import validate_amount_interval


class SleepModule(Module):
    module_name = 'SLEEP_MODULE'
    module_config = 'sleep_module_config'

    def execute(self, sleep_interval, account):
        logging.info("Additional sleeping module")
        randomized_sleeping(sleep_interval)

    def log(self):
        return "SLEEP MODULE"

    def sleep_after(self):
        return False

    def parse_params(self, module_params):
        validate_amount_interval(module_params['sleep_interval'])

        return [module_params['sleep_interval']]



from sybil_engine.data.networks import get_chain_instance
from sybil_engine.utils.app_account_utils import AppAccount
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.linea_day_2_contract import LineaDay2
from web3_wizzard_lib.core.utils.sub_module import SubModule


class LineaCultureDay2(SubModule):
    module_name = 'LINEA_CULTURE_2'

    def execute(self, account: AppAccount, chain='LINEA'):
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)

        linea = LineaDay2('0xB8DD4f5Aa8AD3fEADc50F9d670644c02a07c9374', web3)
        linea.approve(account)

    def log(self):
        return "LINEA CULTURE 2"



from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction

from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/aave.json")


class AaveContract(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def depositETH(self, account, amount: int):
        sender = account.address
        txn_params = self.build_generic_data(sender, set_contract_address=False)

        txn_params['value'] = amount

        return self.contract.functions.depositETH(
            '0x11fCfe756c05AD438e312a7fd934381537D3cFfe',
            account.address,
            0
        ).build_transaction(txn_params)

    @evm_transaction
    def withdrawETH(self, account, amount: int):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)

        return self.contract.functions.withdrawETH(
            '0x11fCfe756c05AD438e312a7fd934381537D3cFfe',
            amount,
            account.address
        ).build_transaction(txn_params)


from loguru import logger
from sybil_engine.domain.balance.balance_utils import interval_to_eth_balance, interval_to_erc20_balance
from sybil_engine.domain.balance.tokens import Erc20Token

from web3_wizzard_lib.core.contract.velocore_lens import VelocoreLensContract
from web3_wizzard_lib.core.contract.velocore_pool import VelocorePoolContract
from web3_wizzard_lib.core.modules.liquidity.pool import Pool


class VelocorePool(Pool):
    module_name = 'VELOCORE_LIQUIDITY'
    pool_contract = 'VELOCORE_POOL'
    pool_lens_contract = 'VELOCORE_LENS'

    def __init__(self, chain_instance, web3):
        super().__init__(chain_instance, web3)
        self.chain_instance = chain_instance
        self.velocore_pool = VelocorePoolContract(self.velocore_pool_contract_address, self.web3)
        self.velocore_lens = VelocoreLensContract('0xaA18cDb16a4DD88a59f4c2f45b5c91d009549e06', self.web3)

    def deposit(self, amount_interval, account, token, min_native_balance, chain):
        if token == 'ETH':
            amount = interval_to_eth_balance(amount_interval, account, chain, self.web3)

            if amount_interval == 'all_balance':
                amount = amount.minus(min_native_balance)
        else:
            amount = interval_to_erc20_balance(amount_interval, account, token, chain, self.web3)

            erc20_token = Erc20Token(self.chain_instance['chain'], "wstETH", self.web3)

            if erc20_token.allowance(account, self.velocore_pool.contract_address) < amount.wei:
                erc20_token.approve(account, self.velocore_pool.contract_address)

        pool_address = '0x2BD146e7d95cea62C89fcCA8E529e06EEc1b053C'

        logger.info(f"Deposit {amount}")

        self.velocore_pool.deposit(account, int(amount.wei * 0.99), token, pool_address)

    def withdraw(self, account, token, chain):
        pool_address = '0x2BD146e7d95cea62C89fcCA8E529e06EEc1b053C'

        gauge = self.velocore_lens.query_gauge(pool_address, account)

        usdc_eth_vlp = gauge[14]
        eth = gauge[9]

        logger.info(f"Withdraw {eth / 10 ** 18} ETH")

        if eth / 10 ** 18 < 0.0005:
            logger.info("Withdraw amount less than 0.0005, skip")
            return

        self.velocore_pool.withdraw(account, eth, usdc_eth_vlp, pool_address)



memory = {

}

def add_value(key, value):
    memory[key] = value

def add_memory_list(key, value):
    memory[key] = value


def accumulate_by_key(key, value):
    if key not in memory:
        memory[key] = []

    memory[key].append(value)
    # if isinstance(value, map):
    #     (list(value))
    # else:
    #     memory[key] += value


def remove_key(key):
    del memory[key]


def get_by_key(key):
    return memory.get(key)



from loguru import logger
from sybil_engine.config.app_config import get_network
from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.domain.balance.balance import Erc20Balance, NativeBalance
from sybil_engine.domain.balance.balance_utils import get_native_balance, interval_to_eth_balance
from sybil_engine.domain.balance.tokens import Erc20Token
from sybil_engine.module.module import Module
from sybil_engine.utils.utils import randomized_sleeping
from sybil_engine.utils.web3_utils import init_web3, get_amount_to_bridge_usdc, get_max_balance_data

from web3_wizzard_lib.core.contract.stargate_farming import StargateFarming
from web3_wizzard_lib.core.contract.stargate_router import StargateRouter
from web3_wizzard_lib.core.modules.stargate.pool import add_to_pool, find_max_pool_chain, get_pool_token_name, get_pool_balances
from web3_wizzard_lib.core.modules.stargate.stargate_balance_utils import StargatePoolToken

all_chains = {
    'MAIN': {
        'USDC': ['ARBITRUM', 'BASE', 'BSC', 'POLYGON', 'AVALANCHE', 'OPTIMISM', 'FANTOM'],
        'ETH': ['ARBITRUM', 'BASE', 'LINEA', 'OPTIMISM']
    },
    'LOCAL': {
        'USDC': ['BASE'],
        'ETH': ['BASE']
    },
    'GOERLI': {
        'USDC': ['LINEA', 'BASE'],
        'ETH': ['LINEA', 'BASE']
    }
}


class StargateFarmingModule(Module):
    module_name = 'STARGATE_FARMING'
    module_config = 'stargate_farming_config'

    def execute(self, bridge_amount_interval, chain, operation, all_balance_mode, token, operation_sleep_interval,
                account):
        try:
            logger.info(f"=============== Stargate staking: {account.address} ===============")

            chains = all_chains[get_network()][token]

            if 'ADD_TO_POOL' in operation:
                self.add_to_pool_logic(account, all_balance_mode, bridge_amount_interval, chain, chains, token)
                randomized_sleeping(operation_sleep_interval)

            if 'DEPOSIT' in operation:
                self.deposit_logic(account, chains, token)
                randomized_sleeping(operation_sleep_interval)

            if 'WITHDRAW' in operation:
                self.withdraw_logic(account, chains, token)
                randomized_sleeping(operation_sleep_interval)

            if 'REDEEM_FROM_POOL' in operation:
                self.redeem_from_pool_logic(account, chains, token)
                randomized_sleeping(operation_sleep_interval)
        except Exception as e:
            logger.info(f"Error during farming: {e}")

    def redeem_from_pool_logic(self, account, chains, token):
        pool_balances = get_pool_balances(token, account, chains)
        chain, pool_balance, web3 = find_max_pool_chain(pool_balances)
        chain_instance = get_chain_instance(chain)
        stargate_router = StargateRouter(get_contracts_for_chain(chain)["STARGATE_ROUTER"], web3)
        logger.info(f"Redeem from {chain} pool {pool_balance}")
        if token == 'USDC':
            stargate_pool_id = chain_instance['stargate_usdc_pool']
        else:
            stargate_pool_id = chain_instance['stargate_eth_pool']
        if pool_balance.wei > 0:
            stargate_router.instant_redeem(account, pool_balance.wei, stargate_pool_id)
        else:
            logger.info(f"Pool is empty, ignoring")

    def withdraw_logic(self, account, chains, token):
        logger.info("Withdraw from farming")
        farm_balances = self.get_farm_balances(account, token, chains)
        chain, farm_balance, web3 = find_max_pool_chain(farm_balances)
        farming_address = get_contracts_for_chain(chain)['STARGATE_FARMING']
        stargate_farming = StargateFarming(farming_address, web3)
        logger.info(f"Balance in farming {farm_balance}, withdrawing all")
        if farm_balance.wei > 0:
            stargate_farming.withdraw(account, farm_balance.wei)
        else:
            logger.info(f"Farming balance is empty, ignoring")

    def deposit_logic(self, account, chains, token):
        logger.info("Deposit to farm")

        pool_balances = get_pool_balances(token, account, chains)
        chain, pool_balance, web3 = find_max_pool_chain(pool_balances)

        chain_instance = get_chain_instance(chain)
        farming_address = get_contracts_for_chain(chain)['STARGATE_FARMING']

        stargate_farming = StargateFarming(farming_address, web3)
        logger.info(f"Amount in {chain} pool {pool_balance}")

        pool_token_name = get_pool_token_name(token)
        stargate_pool_token = StargatePoolToken(chain, pool_token_name, web3)

        if stargate_pool_token.allowance(account, farming_address) < pool_balance.wei:
            stargate_pool_token.approve(account, farming_address)

        stargate_farming.deposit(account, pool_balance)

    def add_to_pool_logic(self, account, all_balance_mode, bridge_amount_interval, chain, chains, token):

        def get_balance_to_pool(token_type, chain_instance, web3):
            if token_type == 'USDC':
                stargate_token = Erc20Token(chain, 'USDC', web3)
                return stargate_token.balance(account)
            else:
                return get_native_balance(account, web3, chain_instance)

        if bridge_amount_interval == 'all_balance':
            if all_balance_mode == 0:
                max_balance_chain, amount_to_pool, native_balance, web3 = get_max_balance_data(token, chains, account)
                chain_instance = get_chain_instance(max_balance_chain)
                amount_to_pool = native_balance
            else:
                chain_instance = get_chain_instance(chain)
                web3 = init_web3(chain_instance, account.proxy)
                amount_to_pool = get_balance_to_pool(token, chain_instance, web3)
            amount_to_pool = amount_to_pool.minus(self.min_native_balance)
        else:
            chain_instance = get_chain_instance(chain)
            web3 = init_web3(chain_instance, account.proxy)
            amount_to_pool = get_amount_to_bridge_usdc(
                bridge_amount_interval) if token == 'USDC' else interval_to_eth_balance(bridge_amount_interval, account,
                                                                                        chain_instance['chain'], web3)

        logger.info(f"Adding to stargate {chain_instance['chain']} pool")
        add_to_pool(account, amount_to_pool, token, chain_instance, web3)

    def get_farm_balances(self, account, token, chains):

        def get_balance_for_chain(chain):
            web3 = init_web3(get_chain_instance(chain), account.proxy)
            farming_address = get_contracts_for_chain(chain)['STARGATE_FARMING']
            farming = StargateFarming(farming_address, web3)

            if token == 'USDC':
                return chain, Erc20Balance(farming.user_info(account), chain, token), web3
            else:
                return chain, NativeBalance(farming.user_info(account), chain, token), web3

        return [get_balance_for_chain(chain) for chain in chains]

    def log(self):
        return "STARGATE FARMING"

    def parse_params(self, module_params):
        return (
            module_params['bridge_amount_interval'],
            module_params['chain'],
            module_params['operation'],
            module_params['all_balance_mode'],
            module_params['token'],
            module_params['operation_sleep_interval']
        )



from sybil_engine.contract.send import Send
from sybil_engine.domain.balance.balance import NativeBalance
from sybil_engine.domain.balance.balance_utils import from_eth_to_wei
from sybil_engine.utils.app_account_utils import AppAccount

from web3_wizzard_lib.core.modules.nft.nft_submodule import NftSubmodule


class LineaCultureWeek3Day4(NftSubmodule):
    module_name = 'LINEA_CULTURE_3_5'
    nft_address = '0x5A77B45B6f5309b07110fe98E25A178eEe7516c1'

    def execute(self, account: AppAccount, chain='LINEA', wei_cost=from_eth_to_wei(0)):
        Send(
            None,
            self.create_web3(account, chain)
        ).send_to_wallet(
            account,
            self.nft_address,
            NativeBalance(wei_cost, chain, "ETH"),
            f"0x731133e9000000000000000000000000{account.address[2:]}0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000100000000000000000000000000000000000000000000000000000000000000800000000000000000000000000000000000000000000000000000000000000000"
        )

    def log(self):
        return "LINEA CULTURE 3 WEEK DAY 5 (Demortal)"



from loguru import logger
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.domain.balance.tokens import Erc20Token
from sybil_engine.module.module import Order, Module
from sybil_engine.utils.accumulator import add_accumulator_balance
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.dmail_send_mail import DmailSend


class Erc20BalanceModule(Module):
    module_name = 'ERC20_BALANCE_MODULE'
    module_config = 'erc20_balance'
    allowed_chains = ['ZKSYNC', 'LINEA', 'SCROLL', 'MANTA', 'BASE']
    cumulative_sum = 0

    def execute(self, token_address, chain, account):
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)

        erc20_token = Erc20Token('LINEA', token_address, web3)

        balance = erc20_token.balance(account)

        self.cumulative_sum = self.cumulative_sum + balance.wei

        logger.info(f"{account.address} balance is {balance}")
        add_accumulator_balance("Total balance is", balance.wei)

    def log(self):
        return "ERC20 BALANCE"

    def get_contract_class(self):
        return DmailSend

    def parse_params(self, module_params):
        self.validate_supported_chain(module_params['chain'])

        return module_params['token_address'], module_params['chain']

    def order(self):
        return Order.RANDOM



import random

from loguru import logger
from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.domain.balance.balance_utils import interval_to_eth_balance, verify_balance, from_wei_to_eth
from sybil_engine.module.module import Module
from sybil_engine.utils.utils import ConfigurationException, randomized_sleeping, print_exception_chain
from sybil_engine.utils.validation_utils import validate_chain, validate_refuel_interval
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.merkly_minter import MerklyMinter
from web3_wizzard_lib.core.modules.stargate.l0_data import l0_chain_ids


class MerklyRefuel(Module):
    module_name = 'MERKLY_REFUEL'
    module_config = 'merkly_refuel_config'

    def execute(self, start_chain, end_chain, chains, refuel_amount_interval, sleeping_interval, account):
        from_chain_instance = get_chain_instance(start_chain)
        web3 = init_web3(from_chain_instance, account.proxy)

        if start_chain not in chains or end_chain not in chains:
            raise ConfigurationException(f"'end_chain' and 'start_start' should be in 'chains'. Actual:"
                                         f"start_chain: {start_chain} end_chain: {end_chain} chains: {chains}"
                                         )

        merkly_contract = get_contracts_for_chain(start_chain)['MERKLY']
        merkly = MerklyMinter(merkly_contract, web3)

        self.shuffle_chain_sequence(start_chain, end_chain, chains)

        logger.info(f"Refuel chain sequence: {chains}")

        for current_chain, next_chain in zip(chains, chains[1:]):
            native_without_min = verify_balance(self.min_native_balance, from_chain_instance, account, web3)

            amount = interval_to_eth_balance(refuel_amount_interval, account, start_chain, web3)

            if amount.wei > native_without_min.wei:
                amount = native_without_min

            logger.info(f"Refuel {current_chain} > {next_chain} | {amount}")

            hex_amount = merkly.decimal_to_hex_padded(amount.wei)

            try:
                fee = merkly.estimate_send_fee(account, l0_chain_ids[next_chain], hex_amount)
                logger.info(f"L0 Fee: {from_wei_to_eth(fee)} ETH")

                merkly.refuel(account, l0_chain_ids[next_chain], amount.wei, fee)
                randomized_sleeping(sleeping_interval)
            except Exception as e:
                print_exception_chain(e)

    def shuffle_chain_sequence(self, start_chain, end_chain, chain_sequence):
        if start_chain != '':
            chain_sequence.remove(start_chain)

        if end_chain != '' and end_chain in chain_sequence:
            chain_sequence.remove(end_chain)

        random.shuffle(chain_sequence)

        if start_chain != '':
            chain_sequence.insert(0, start_chain)

        if end_chain != '':
            chain_sequence.append(end_chain)

        return chain_sequence

    def log(self):
        return "MERKLY REFUEL"

    def parse_params(self, module_params):
        validate_chain(module_params['end_chain'])
        validate_refuel_interval(module_params['refuel_amount_interval'])

        return (
            module_params['start_chain'],
            module_params['end_chain'],
            module_params['chains'],
            module_params['refuel_amount_interval'],
            module_params['sleeping_interval']
        )



from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/octomos.json")


class OctomosContract(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def launchpadBuy(self,
                     account,
                     undef,
                     launchpadId,
                     slotId,
                     quantity,
                     additional,
                     data
                     ):
        txn_params = self.build_generic_data(account.address, False)

        return self.contract.functions.launchpadBuy(
            undef,
            launchpadId,
            slotId,
            quantity,
            additional,
            data
        ).build_transaction(txn_params)



import time
from random import randint

from eth_account.messages import encode_defunct, SignableMessage
from fake_useragent import UserAgent
from loguru import logger

import requests
from sybil_engine.utils.utils import randomized_sleeping

url = "https://api.rabby.io/v2/points/claim_snapshot"

headers = {
    "Host": "api.rabby.io",
    "Connection": "keep-alive",
    "Content-Length": "224",
    "sec-ch-ua": '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
    "X-Version": "0.92.48",
    "sec-ch-ua-mobile": "?0",
    "User-Agent": UserAgent().random,
    "x-api-ts": str(int(time.time())),
    "Content-Type": "application/json",
    "x-api-ver": "v2",
    "Accept": "application/json, text/plain, */*",
    "X-Client": "Rabby",
    "sec-ch-ua-platform": '"Windows"',
    "Origin": "chrome-extension://acmacodkjb dgmoleebolmdjonilkdbch",
    "Sec-Fetch-Site": "none",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Dest": "empty",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "uk-UA,uk;q=0.9,en-US;q=0.8,en;q=0.7",
}


def claimer_rabby(account, web3):
    try:
        address = account.address.lower()
        msg = f"{address} Claims Rabby Points"

        payload = {
            "id": address,
            "signature": get_signed_code(web3, account, msg),
            "invite_code": "FREEBONUS"
        }

        proxies = {}
        proxies.update(
            {"http": account.proxy,
            "https": account.proxy}
        )

        response = make_request_with_retries(url, headers=headers, json=payload, proxies=proxies)

        if response.json().get("error_code") == 0:
            logger.success(f"{address} | Claimed!")
        else:
            resp_msg = response.json().get("error_msg")
            logger.info(f"{address} | {resp_msg} | {response.json()}")
    except Exception as e:
        logger.error(f"{account}... | {e}")


def sign(web3, account, encoded_msg: SignableMessage):
    return web3.eth.account.sign_message(encoded_msg, account.key)


def get_signed_code(web3, account, msg) -> str:
    return sign(web3, account, encode_defunct(text=msg)).signature.hex()

def make_request_with_retries(url, headers, json, proxies, max_retries=5):
    for attempt in range(max_retries):
        response = requests.post(url, headers=headers, json=json, proxies=proxies)
        if response.status_code == 429:
            retry_after = int(response.headers.get("Retry-After", randint(1, 5)))
            logger.info(f"Rate limited. Retrying after {retry_after} seconds. Attempt {attempt + 1}/{max_retries}")
            time.sleep(retry_after)
        else:
            return response
    logger.error("Max retries reached. Giving up.")
    return None


from loguru import logger
from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.tavaera_id import TavaeraID
from web3_wizzard_lib.core.contract.tavaera_mint import Tavaera
from web3_wizzard_lib.core.utils.sub_module import SubModule


class TavaeraModule(SubModule):
    module_name = 'TAVAERA'

    def execute(self, account, chain='ZKSYNC'):
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)

        chain_contracts = get_contracts_for_chain(chain)

        tavaera_id = TavaeraID(chain_contracts['TAVAERA_ID'], web3)
        tavaera = Tavaera(chain_contracts['TAVAERA_MINT'], web3)

        logger.info("Mint Tavaera ID")
        tavaera_id.mint_citizen_id(account)

        logger.info("Mint Tavaera NFT")
        tavaera.mint(account)

    def log(self):
        return "CREATE TAVAERA ACCOUNT"



import random

from loguru import logger
from sybil_engine.contract.erc20contract import Erc20Contract
from sybil_engine.contract.send import Send
from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.data.tokens import get_tokens_for_chain
from sybil_engine.domain.balance.balance import NotEnoughNativeBalance
from sybil_engine.domain.balance.balance_utils import interval_to_erc20_balance, \
    interval_to_native_balance, verify_balance
from sybil_engine.utils.utils import ConfigurationException
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.orbiter_starknet import OrbiterStarknet
from web3_wizzard_lib.core.modules.bridge.bridge import Bridge

bridge_codes = {
    "ETH": 9001,
    "ARBITRUM": 9002,
    "POLYGON": 9006,
    "OPTIMISM": 9007,
    "BSC": 9015,
    "ARBITRUM_NOVA": 9016,
    "ZKSYNC": 9014,
    "POLYGON_ZK": 9017,
    "BASE": 9021,
    "LINEA": 9023,
    "SCROLL": 9019,
    #"X_LAYER": 9027, DONT USE
    "ZORA": 9030,
    "MANTA": 9031,
    "ZKFAIR": 9038
}


class OrbiterSubModule(Bridge):
    supported_tokens = ['NATIVE', 'USDC', 'USDT', 'WETH']

    def execute(self, bridge_amount_interval, from_chain, to_chain, token, min_native_balance, account):
        from_chain_instance = get_chain_instance(from_chain)
        web3 = init_web3(from_chain_instance, account.proxy)

        if isinstance(to_chain, list):
            to_chain = random.choice(to_chain)
        if to_chain not in bridge_codes:
            raise ConfigurationException(f"{to_chain} not supported in orbiter")

        amount = self.__calculate_bridge_amount(account, bridge_amount_interval, from_chain_instance, to_chain, token,
                                                min_native_balance, web3)

        logger.info(f"Bridge {from_chain} –> {to_chain} | {amount}")
        self.__perform_bridge(account, amount, from_chain, to_chain, token, web3)

    def __perform_bridge(self, account, amount, from_chain, to_chain, token, web3):
        if token == 'NATIVE':
            self.__bridge_native(account, amount, from_chain, to_chain, web3)
        else:
            self.__bridge_token(account, amount, from_chain, web3)

    def __calculate_bridge_amount(self, account, bridge_amount_interval, from_chain_instance, to_chain, token,
                                  min_native_balance, web3):
        if token == 'NATIVE':
            native_without_min = verify_balance(min_native_balance, from_chain_instance, account, web3)

            amount = interval_to_native_balance(bridge_amount_interval, account, from_chain_instance['chain'], web3)
            if amount.wei > native_without_min.wei:
                if bridge_amount_interval == 'all_balance':
                    amount = native_without_min
                else:
                    raise NotEnoughNativeBalance(
                        f"The account balance ({native_without_min}) < bridging amount ({amount}).")
        else:
            amount = interval_to_erc20_balance(bridge_amount_interval, account, token, from_chain_instance['chain'],
                                               web3)
            if bridge_amount_interval == 'all_balance':
                amount.wei = amount.wei - 10 ** 5

        amount.wei = self.__adjust_amount_by_bridge_code(amount.wei, to_chain)

        return amount

    def __adjust_amount_by_bridge_code(self, wei_amount, to_chain):
        bridge_code = bridge_codes.get(to_chain)
        if not bridge_code:
            raise ConfigurationException(f"No bridge code found for {to_chain}")
        return int(f"{wei_amount // 10 ** 4}{bridge_code}")

    def __bridge_native(self, account, amount, from_chain, to_chain, web3):
        if to_chain == 'STARKNET':
            if amount.readable() <= 0.0065 or amount.readable() >= 5:
                raise NotEnoughNativeBalance(f"Limit range amount for starknet 0.0065 – 5 ETH | {amount}")

            if from_chain != 'ZKSYNC':
                raise ConfigurationException("only ZKSYNC can bridge to STARKNET")

            orbiter_wallet = get_contracts_for_chain(from_chain)['ORBITER_STARKNET']
            orbiter_starknet = OrbiterStarknet(orbiter_wallet, web3)
            orbiter_starknet.bridge(account, amount)
        else:
            orbiter_wallet = get_contracts_for_chain(from_chain)['ORBITER']
            send = Send(orbiter_wallet, web3)
            send.send_to_wallet(account, orbiter_wallet, amount)

    def __bridge_token(self, account, amount, from_chain, web3):
        contracts = get_contracts_for_chain(from_chain)

        if 'ORBITER_ERC20' not in contracts:
            raise ConfigurationException(f"ERC20 is not supported for {from_chain}")

        erc20 = Erc20Contract(get_tokens_for_chain(from_chain)[amount.token], web3)
        erc20.transfer(account, amount, contracts['ORBITER_ERC20'])



from sybil_engine.data.contracts import get_contracts_for_chain


class Pool:
    pool_contract = ''

    def __init__(self, chain_instance, web3):
        self.chain_contracts = get_contracts_for_chain(chain_instance['chain'])
        self.velocore_pool_contract_address = self.chain_contracts[self.pool_contract]
        self.web3 = web3

    def deposit(self, amount_interval, account, token, min_native_balance, chain):
        pass

    def withdraw(self, account, token, chain):
        pass



from datetime import datetime

from eth_account.messages import encode_defunct
from sybil_engine.contract.contract import Contract
from sybil_engine.contract.erc20contract import MAX_ALLOWANCE
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.data.tokens import get_tokens_for_chain
from sybil_engine.utils.file_loader import load_abi
from web3 import Web3

abi = load_abi("resources/abi/syncswap_router.json")


class SyncSwapRouter(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def swap(self, account, amount_to_swap, token_in_data_address, token_in_address, pool_address, amount_out_min,
             withdraw_mode=b'\x01'):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)
        txn_params['value'] = amount_to_swap.wei if amount_to_swap.token == 'ETH' else 10000000000000

        steps = self.steps(amount_to_swap, pool_address, sender, token_in_address, token_in_data_address,
                           withdraw_mode)

        paths = [{"steps": steps, "tokenIn": token_in_address, "amountIn": amount_to_swap.wei}]
        deadline = int(datetime.now().timestamp() + 60 * 60 * 3)

        return self.contract.functions.swap(paths, amount_out_min, deadline).build_transaction(txn_params)

    def steps(self, amount_to_swap, pool_address, sender, token_in_address, token_in_data_address, withdraw_mode):
        if isinstance(pool_address, str):
            data = b'\x00' * 12 + bytes.fromhex(token_in_data_address[2:]) + b'\x00' * 12 + bytes.fromhex(
                sender[2:]) + b'\x00' * 31 + withdraw_mode

            steps = [
                {
                    "pool": pool_address,
                    "data": data,
                    "callback": get_tokens_for_chain(amount_to_swap.chain)['ZERO_ADDRESS'],
                    "callbackData": b''
                }
            ]
        else:
            usdc_address = get_tokens_for_chain(amount_to_swap.chain)['USDC']

            pool_pairs = list(pool_address.items())
            if amount_to_swap.token == 'USDT':
                zero_pool = pool_pairs[0][1]  # USDC
                first_pool = pool_pairs[1][1]  # USDT

                data_fir_step = b'\x00' * 12 + bytes.fromhex(token_in_address[2:]) + b'\x00' * 12 + bytes.fromhex(
                    sender[2:]) + b'\x00' * 31 + b'\x00'

                data_sec_step = b'\x00' * 12 + bytes.fromhex(usdc_address[2:]) + b'\x00' * 12 + bytes.fromhex(
                    sender[2:]) + b'\x00' * 31 + withdraw_mode
            else:
                zero_pool = pool_pairs[1][1]  # USDT
                first_pool = pool_pairs[0][1]  # USDC

                data_fir_step = b'\x00' * 12 + bytes.fromhex(
                    token_in_data_address[2:]) + b'\x00' * 12 + bytes.fromhex(zero_pool[2:]) + b'\x00' * 31 + b'\x00'

                data_sec_step = b'\x00' * 12 + bytes.fromhex(usdc_address[2:]) + b'\x00' * 12 + bytes.fromhex(
                    sender[2:]) + b'\x00' * 31 + b'\x02'

            steps = [
                {
                    "pool": first_pool,
                    "data": data_fir_step,
                    "callback": get_tokens_for_chain(amount_to_swap.chain)['ZERO_ADDRESS'],
                    "callbackData": b''
                },
                {
                    "pool": zero_pool,
                    "data": data_sec_step,
                    "callback": get_tokens_for_chain(amount_to_swap.chain)['ZERO_ADDRESS'],
                    "callbackData": b''
                }
            ]

        return steps

    @evm_transaction
    def add_liquidity2(self, account, amount):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)

        txn_params['value'] = amount.wei

        return self.contract.functions.addLiquidity2(
            '0x814A23B053FD0f102AEEda0459215C2444799C70',
            [
                (
                    '0x0000000000000000000000000000000000000000',
                    amount.wei
                )
            ],
            f'0x000000000000000000000000{account.address[2:]}',
            0,
            '0x0000000000000000000000000000000000000000',
            b'0x'
        ).build_transaction(txn_params)

    @evm_transaction
    def burn_liquidity(self, account, pool, amount):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)

        return self.contract.functions.burnLiquiditySingle(
            pool,
            amount.wei,  # liqudity
            Web3.to_bytes(
                hexstr=f'0x0000000000000000000000005300000000000000000000000000000000000004000000000000000000000000{account.address[2:]}0000000000000000000000000000000000000000000000000000000000000001'),
            amount.wei,  # minAmount
            Web3.to_bytes(hexstr='0x0000000000000000000000000000000000000000'),
            b'0x'
        ).build_transaction(txn_params)
